from typing import Callable, Any, Tuple, Dict, Awaitable

from pydantic import BaseModel, Field


class Task(BaseModel):
    func: Callable[..., Awaitable[Any]]
    args: Tuple[Any, ...] = Field(default_factory=tuple)
    kwargs: Dict[str, Any] = Field(default_factory=dict)


import asyncio
from asyncio import Queue, AbstractEventLoop
from typing import Optional
from threading import Event

from ._task import Task


class TaskQueue:
    """
    An asyncio.Queue confined to a single (owning) event loop, but usable
    from *any* loop or thread via cross-loop marshaling.
    """

    def __init__(self) -> None:
        self._loop: Optional[AbstractEventLoop] = None
        self._q: Optional[Queue[Task]] = None
        self._started = Event()
        self._stopped = Event()

    async def start(self) -> None:
        """
        Initialize the queue inside the running (owning) loop.
        Must be awaited from the target loop/thread that will own the queue.
        """
        self._loop = asyncio.get_running_loop()
        self._q = asyncio.Queue()
        self._started.set() # unlock the queue

    # ----------------------------
    # Async API usable from ANY loop
    # ----------------------------

    async def put(self, item: Task) -> None:
        """
        Works when called from the owning loop OR any other event loop.
        If called from a foreign loop, the operation is marshaled to the owning loop.
        """
        self._started.wait() # secure that the queue is started
        assert self._loop is not None and self._q is not None
        # Foreign loop: marshal to owning loop and await the cross-loop future.
        fut = asyncio.run_coroutine_threadsafe(self._q.put(item), self._loop)
        await asyncio.wrap_future(fut)

    async def get(self) -> Task:
        """
        Works when called from the owning loop OR any other event loop.
        If called from a foreign loop, the operation is marshaled to the owning loop.
        """
        self._started.wait() # secure that the queue is started
        assert self._loop is not None and self._q is not None
        fut = asyncio.run_coroutine_threadsafe(self._q.get(), self._loop)
        return await asyncio.wrap_future(fut)


from functools import wraps
from typing import Callable, Awaitable, Any

from ._task import Task
from ._task_queue import TaskQueue
from ._worker import Worker
from ._head import HeadWorker


class WorkerPool:

    def __init__(self, concurrency: int):
        """
        The main worker pool class.
        :param concurrency: The number of concurrency workers.
        """

        # initialize the worker queue
        self._worker_queue = TaskQueue()

        # initialize the worker
        self._head_worker = HeadWorker(self._worker_queue)
        self._workers = [Worker(self._worker_queue) for _ in range(concurrency)]

    def start(self) -> None:
        """
        Start the worker pool.
        :return: None
        """
        self._head_worker.start()
        for worker in self._workers:
            worker.start()

    def join(self) -> None:
        """
        Block until all workers are done.
        :return: None
        """
        self._head_worker.join()
        for worker in self._workers:
            worker.join()

    def task(self, func: Callable[..., Awaitable[Any]]):
        """
        Task decorator.
        :param func: The function to be executed.
        :return: None
        """

        @wraps(func)
        async def wrapper(*args, **kwargs):
            # create the task object and put it in the execution queue
            task = Task(func=func, args=args, kwargs=kwargs)
            await self._worker_queue.put(task)
            return None

        return wrapper



from apscheduler.schedulers.background import BackgroundScheduler


class Scheduler:

    def __init__(self):
        self._scheduler = BackgroundScheduler()

    def start(self):
        self._scheduler.start()


from fastapi import FastAPI
from ._ping import ping_router


def create_rest_app(title:str="Ventricle",root_path: str = "") -> FastAPI:
    """
    Create and configure a FastAPI REST application.
    :param title: The title of the application.
    :param root_path: The root path of the application.
    :return: The configured FastAPI REST application.
    """
    app = FastAPI(title=title,root_path=root_path)

    app.include_router(ping_router)

    return app


from apscheduler.schedulers.background import BackgroundScheduler

def create_schedular_app() -> BackgroundScheduler:
    schedular = BackgroundScheduler() # todo move to custom scheduler class

    return schedular


import asyncio
from ._task_queue import TaskQueue
from ._worker import Worker


class HeadWorker(Worker):
    """
        Coordinating Worker Thread. Owns the managing event loop.
    """

    def __init__(self, queue: TaskQueue) -> None:
        super().__init__(queue)

    async def _run_async(self):
        await self.queue.start()
        await asyncio.Event().wait()
